import os.path
from pathlib import Path
from shutil import copyfile
import cv2

class RenderResult:
    """Result of any form of render.

    Treated as a series of paths to the output files. Handles different output types, cameras, and frame counts.
    """

    def __init__(self, render_paths: dict[tuple, str]):
        """Expects dictionary of:

        (output_key, camera_name, frame_number) -> Path.
        """

        self._render_paths = render_paths

        self.output_types = list(set([k[0] for k in render_paths.keys()]))
        self.camera_names = list(set([k[1] for k in render_paths.keys()]))
        self.frames = list(set([k[2] for k in render_paths.keys()]))

    @property
    def num_cameras(self) -> int:
        return len(self.camera_names)

    @property
    def num_frames(self) -> int:
        return len(self.frames)

    def get_render_path(self, output_type: str, camera_name: str = None, frame_number: int = None) -> str:

        if camera_name is None:
            if self.num_cameras != 1:
                raise ValueError("Must specify camera name if more than one camera. Cameras: ", self.camera_names)

            camera_name = self.camera_names[0]

        if frame_number is None:
            if self.num_frames != 1:
                raise ValueError("Must specify frame number if more than one frame. Frames: ", self.frames)

            frame_number = self.frames[0]

        return self._render_paths[(output_type, camera_name, frame_number)]

    def save_all(self, output_directory: str, suppress_camera_name: bool = True, suppress_frame_number: bool = True):
        """Save all the files to a directory.

        By default, saves files in format {data_type}_{camera_name}_{frame_number}.{ext}.

        If only 1 camera or only 1 frame, those formats will be suppressed by default.
        """

        os.makedirs(output_directory, exist_ok=True)

        suppress_camera_name = suppress_camera_name and self.num_cameras == 1
        suppress_frame_number = suppress_frame_number and self.num_frames == 1

        for (data_type, camera, frame), path in self._render_paths.items():

            fname = f'{data_type}'
            if not suppress_camera_name:
                fname += f'_{camera}'
            if not suppress_frame_number:
                fname += f'_{frame:06d}'

            fname += os.path.splitext(path)[1]

            copyfile(path, os.path.join(output_directory, fname))

    def save_file(self, output_path: str, output_type: str, camera_name: str = None, frame_number: int = None):
        """Save a single file to a path."""

        if os.path.dirname(output_path):
            os.makedirs(os.path.dirname(output_path), exist_ok=True)

        path = self.get_render_path(output_type, camera_name, frame_number)

        target_ext = os.path.splitext(output_path)[1]
        rendered_ext = os.path.splitext(path)[1]
        if target_ext != rendered_ext:
            raise ValueError(f"Output path must have same extension as rendered path. Got {target_ext}, expected {rendered_ext}.")

        copyfile(path, output_path)



import tempfile

import blendersynth as bsyn
import unittest
import cv2
import numpy as np

class UnitTestCompositor(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        bsyn.render.set_cycles_samples(10)
        bsyn.render.set_resolution(256, 256)

        bsyn.world.set_color((0.5, 1.0, 0.5)) # Set color to test for color mapping.

        cube = bsyn.Mesh.from_primitive('cube')
        camera = bsyn.Camera()
        comp = bsyn.Compositor()

        camera.location = (0, 0, 5)
        camera.look_at_object(cube)

        normal_aov = bsyn.aov.NormalsAOV()
        cube.assign_aov(normal_aov)

        comp.define_output('Image', 'rgb', is_data=False)
        comp.define_output(normal_aov, 'normal8', is_data=True, color_depth="8")
        comp.define_output(normal_aov, 'normal16', is_data=True, color_depth="16")
        cls.render_result = comp.render()

        # Load images.
        cls.image_data = {}
        for key in ['rgb', 'normal8', 'normal16']:
            path = cls.render_result.get_render_path(key)
            cls.image_data[key] = cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB)

    def test_normal_8_render(self):
        img = self.image_data['normal8']

        # Normal should point in camera's negative Z.
        self.assertSequenceEqual(list(img[128, 128]), [128, 128, 0])

    def test_normal_16_render(self):
        img = self.image_data['normal16']

        # Normal should point in camera's negative Z.
        self.assertSequenceEqual(list(img[128, 128]), [128, 128, 0])

    def test_rgb_render(self):
        """Test that an 'Image' render matches the colorspace of the actual output render."""
        img = self.image_data['rgb']

        # Reading Render Result fails, so must save to disk first.
        with tempfile.NamedTemporaryFile(suffix='.png', delete=False) as f:
            bsyn.data.images['Render Result'].save(filepath=f.name)
            render_np = cv2.cvtColor(cv2.imread(f.name), cv2.COLOR_BGR2RGB)

        # All pixels should be the same
        np.testing.assert_allclose(img, render_np, atol=0.1)


"""Compositor node groups for converting sRGB to linear RGB."""

import bpy
from .node_arranger import tidy_tree
from .node_group import tree_add_socket

_srgb_to_linear_single_value_name = "sRGB to Linear Value"
_srgb_to_linear_name = "sRGB to Linear"

def _srgb_to_linear_single_value() -> bpy.types.CompositorNodeGroup:
    """Create node group to go from a single value, sRGB to linear RGB."""

    if _srgb_to_linear_single_value_name in bpy.data.node_groups:
        return bpy.data.node_groups[_srgb_to_linear_single_value_name]

    group = bpy.data.node_groups.new(type="CompositorNodeTree", name=_srgb_to_linear_single_value_name)

    nodes = group.nodes
    input_node = group.nodes.new("NodeGroupInput")
    output_node = group.nodes.new("NodeGroupOutput")

    tree_add_socket(group, "NodeSocketFloat", "sRGB", "INPUT")
    tree_add_socket(group, "NodeSocketFloat", "Linear", "OUTPUT")

    # If sRGB < 0.04045, then Linear = sRGB / 12.92
    # Else, Linear = ((sRGB + 0.055) / 1.055) ^ 2.4

    linear_mode = nodes.new("CompositorNodeMath")
    linear_mode.operation = "DIVIDE"
    linear_mode.inputs[1].default_value = 12.92

    nonlinear_1 = nodes.new("CompositorNodeMath")
    nonlinear_1.operation = "ADD"
    nonlinear_1.inputs[1].default_value = 0.055

    nonlinear_2 = nodes.new("CompositorNodeMath")
    nonlinear_2.operation = "DIVIDE"
    nonlinear_2.inputs[1].default_value = 1.055

    nonlinear_3 = nodes.new("CompositorNodeMath")
    nonlinear_3.operation = "POWER"
    nonlinear_3.inputs[1].default_value = 2.4

    choose_region_node = nodes.new("CompositorNodeMath")
    choose_region_node.operation = "LESS_THAN"
    choose_region_node.inputs[1].default_value = 0.04045

    switch_node = nodes.new("CompositorNodeMixRGB")

    links = group.links

    links.new(input_node.outputs[0], choose_region_node.inputs[0])
    links.new(choose_region_node.outputs[0], switch_node.inputs[0])

    links.new(input_node.outputs[0], nonlinear_1.inputs[0])
    links.new(nonlinear_1.outputs[0], nonlinear_2.inputs[0])
    links.new(nonlinear_2.outputs[0], nonlinear_3.inputs[0])
    links.new(nonlinear_3.outputs[0], switch_node.inputs[1])

    links.new(input_node.outputs[0], linear_mode.inputs[0])
    links.new(linear_mode.outputs[0], switch_node.inputs[2])

    links.new(switch_node.outputs[0], output_node.inputs[0])

    tidy_tree(group)
    return group

def srgb_to_linear() -> bpy.types.CompositorNodeGroup:

    if _srgb_to_linear_name in bpy.data.node_groups:
        return bpy.data.node_groups[_srgb_to_linear_name]

    single_value_group = _srgb_to_linear_single_value()

    group = bpy.data.node_groups.new(type="CompositorNodeTree", name=_srgb_to_linear_name)
    nodes = group.nodes

    input_node = group.nodes.new("NodeGroupInput")
    output_node = group.nodes.new("NodeGroupOutput")

    # Split to colour, then pass each colour through instance of srgb_to_linear, then combine and output.
    tree_add_socket(group, "NodeSocketColor", "sRGB", "INPUT")
    tree_add_socket(group, "NodeSocketColor", "Linear", "OUTPUT")

    splitter = nodes.new("CompositorNodeSepRGBA")
    combiner = nodes.new("CompositorNodeCombRGBA")

    convert_R = nodes.new(type="CompositorNodeGroup")
    convert_G = nodes.new(type="CompositorNodeGroup")
    convert_B = nodes.new(type="CompositorNodeGroup")

    convert_R.node_tree = single_value_group
    convert_G.node_tree = single_value_group
    convert_B.node_tree = single_value_group

    links = group.links
    links.new(input_node.outputs[0], splitter.inputs[0])
    links.new(splitter.outputs[0], convert_R.inputs[0])
    links.new(splitter.outputs[1], convert_G.inputs[0])
    links.new(splitter.outputs[2], convert_B.inputs[0])

    links.new(convert_R.outputs[0], combiner.inputs[0])
    links.new(convert_G.outputs[0], combiner.inputs[1])
    links.new(convert_B.outputs[0], combiner.inputs[2])

    links.new(splitter.outputs[3], combiner.inputs[3]) # Alpha unchanged.

    links.new(combiner.outputs[0], output_node.inputs[0])

    tidy_tree(group)
    return group





from .shader_node_groups import DeformedGeneratedTextureCoordinates
from .node_arranger import tidy_tree
from .node_group import CompositorNodeGroup, ShaderNodeGroup, tree_add_socket
from .sRGB import srgb_to_linear